@@ -39,6 +39,7 @@
  */
 
 #include "grl-plugin-registry.h"
+#include "grl-plugin-registry-priv.h"
 #include "grl-media-plugin-priv.h"
 #include "grl-log.h"
 #include "grl-error.h"
@@ -52,6 +53,10 @@ GRL_LOG_DOMAIN(plugin_registry_log_domain);
 
 #define XML_ROOT_ELEMENT_NAME "plugin"
 
+#define GRL_PLUGIN_INFO_SUFFIX "xml"
+
+#define GRL_PLUGIN_INFO_MODULE "module"
+
 #define GRL_PLUGIN_REGISTRY_GET_PRIVATE(object)                 \
   (G_TYPE_INSTANCE_GET_PRIVATE((object),                        \
                                GRL_TYPE_PLUGIN_REGISTRY,        \
@@ -60,15 +65,28 @@ GRL_LOG_DOMAIN(plugin_registry_log_domain);
 struct _GrlPluginRegistryPrivate {
   GHashTable *configs;
   GHashTable *plugins;
+  GHashTable *plugin_infos;
   GHashTable *sources;
   GHashTable *related_keys;
   GParamSpecPool *system_keys;
   GHashTable *ranks;
   GSList *plugins_dir;
+  GSList *allowed_plugins;
+  gboolean all_plugin_info_loaded;
 };
 
 static void grl_plugin_registry_setup_ranks (GrlPluginRegistry *registry);
 
+static GList *grl_plugin_registry_load_plugin_info_directory (GrlPluginRegistry *registry,
+                                                              const gchar *path,
+                                                              GError **error);
+
+static GrlPluginInfo *grl_plugin_registry_load_plugin_info (GrlPluginRegistry *registry,
+                                                            const gchar *plugin_id,
+                                                            const gchar *file);
+
+static void grl_plugin_registry_load_plugin_info_all (GrlPluginRegistry *registry);
+
 /* ================ GrlPluginRegistry GObject ================ */
 
 enum {
@@ -126,9 +144,11 @@ grl_plugin_registry_init (GrlPluginRegistry *registry)
   registry->priv = GRL_PLUGIN_REGISTRY_GET_PRIVATE (registry);
 
   registry->priv->configs =
-    g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_object_unref);
+    g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
   registry->priv->plugins =
     g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);
+  registry->priv->plugin_infos =
+    g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);
   registry->priv->sources =
     g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
   registry->priv->related_keys =
@@ -153,15 +173,15 @@ config_plugin_rank (GrlPluginRegistry *registry,
 }
 
 static void
-set_plugin_rank (GrlPluginRegistry *registry, GrlPluginDescriptor *plugin)
+set_plugin_rank (GrlPluginRegistry *registry, GrlPluginInfo *info)
 {
-  plugin->info.rank =
+  info->rank =
     GPOINTER_TO_INT (g_hash_table_lookup (registry->priv->ranks,
-					  plugin->info.id));
-  if (!plugin->info.rank) {
-    plugin->info.rank = GRL_PLUGIN_RANK_DEFAULT;
+					  info->id));
+  if (!info->rank) {
+    info->rank = GRL_PLUGIN_RANK_DEFAULT;
   }
-  GRL_DEBUG ("Plugin rank '%s' : %d", plugin->info.id, plugin->info.rank);
+  GRL_DEBUG ("Plugin rank '%s' : %d", info->id, info->rank);
 }
 
 static void
@@ -249,6 +269,7 @@ get_info_from_plugin_xml (const gchar *xml_path)
     info_node = info_node->next;
   }
   if (!info_node) {
+    xmlFreeDoc (doc_ptr);
     return NULL;
   }
 
@@ -266,7 +287,177 @@ get_info_from_plugin_xml (const gchar *xml_path)
   return hash_table;
 }
 
-/* ================ API ================ */
+static GrlPluginInfo *
+grl_plugin_registry_load_plugin_info (GrlPluginRegistry *registry,
+                                      const gchar* plugin_id,
+                                      const gchar *file)
+{
+  GHashTable *info;
+  GrlPluginInfo *plugin_info;
+  gchar *library_filename;
+  gchar *path;
+  gchar *plugin_name;
+
+  /* Load plugin information */
+  info = get_info_from_plugin_xml (file);
+
+  if (!info) {
+    GRL_WARNING ("Invalid information file for '%s' plugin",
+                 plugin_id);
+    return NULL;
+  }
+
+  /* Build plugin library filename */
+  plugin_name = g_hash_table_lookup (info, GRL_PLUGIN_INFO_MODULE);
+  if (!plugin_name) {
+    GRL_WARNING ("Information about '%s' plugin has no reference to module; skipping",
+                 plugin_id);
+    g_hash_table_unref (info);
+    return NULL;
+  }
+  plugin_name = g_strconcat (plugin_name, "." G_MODULE_SUFFIX, NULL);
+  path = g_path_get_dirname (file);
+  library_filename = g_build_filename (path, plugin_name, NULL);
+  g_free (plugin_name);
+  g_free (path);
+
+  /* Build Plugin Info */
+  plugin_info = g_new0 (GrlPluginInfo, 1);
+  plugin_info->id = g_strdup (plugin_id);
+  plugin_info->filename = library_filename;
+  plugin_info->optional_info = info;
+
+  /* Set rank */
+  set_plugin_rank (registry, plugin_info);
+
+  return plugin_info;
+}
+
+static GList *
+grl_plugin_registry_load_plugin_info_directory (GrlPluginRegistry *registry,
+                                                const gchar *path,
+                                                GError **error)
+{
+  GDir *dir;
+  GrlPluginInfo *plugin_info;
+  const gchar *entry;
+  gchar *file;
+  gchar *id;
+  gchar *suffix;
+  GList *loaded_infos = NULL;
+
+  dir = g_dir_open (path, 0, NULL);
+  if (!dir) {
+    GRL_WARNING ("Could not open plugin directory: '%s'", path);
+    g_set_error (error,
+                 GRL_CORE_ERROR,
+                 GRL_CORE_ERROR_LOAD_PLUGIN_FAILED,
+                 "Failed to open plugin directory '%s'", path);
+    return NULL;
+  }
+
+  while ((entry = g_dir_read_name (dir)) != NULL) {
+    if ((suffix = g_strrstr (entry, "." GRL_PLUGIN_INFO_SUFFIX)) != NULL) {
+      file = g_build_filename (path, entry, NULL);
+      id = g_strndup (entry, suffix - entry);
+      /* Skip plugin info if it is already loaded */
+      if (g_hash_table_lookup (registry->priv->plugin_infos, id)) {
+        GRL_DEBUG ("Information about '%s' plugin already loaded; skipping",
+                   id);
+        g_free (id);
+        g_free (file);
+        continue;
+      }
+      /* Check if plugin is allowed or not */
+      if (registry->priv->allowed_plugins &&
+          !g_slist_find_custom (registry->priv->allowed_plugins,
+                                id,
+                                (GCompareFunc) g_strcmp0)) {
+        GRL_DEBUG ("'%s' plugin not allowed; skipping", id);
+        continue;
+      }
+      plugin_info = grl_plugin_registry_load_plugin_info (registry, id, file);
+      g_free (id);
+      g_free (file);
+
+      g_hash_table_insert (registry->priv->plugin_infos,
+                           plugin_info->id,
+                           plugin_info);
+      loaded_infos = g_list_append (loaded_infos, plugin_info);
+    }
+  }
+
+  g_dir_close (dir);
+  return loaded_infos;
+}
+
+static void
+grl_plugin_registry_load_plugin_info_all (GrlPluginRegistry *registry)
+{
+  GSList *plugin_dir;
+  GList *loaded_plugins;
+
+  for (plugin_dir = registry->priv->plugins_dir;
+       plugin_dir;
+       plugin_dir = g_slist_next (plugin_dir)) {
+    loaded_plugins =
+      grl_plugin_registry_load_plugin_info_directory (registry,
+                                                      plugin_dir->data,
+                                                      NULL);
+    g_list_free (loaded_plugins);
+  }
+}
+
+static gboolean
+grl_plugin_registry_load_list (GrlPluginRegistry *registry,
+                               GList *plugin_info_list)
+{
+  GrlPluginInfo *pinfo;
+  gboolean loaded_one = FALSE;
+
+  while (plugin_info_list) {
+    pinfo = (GrlPluginInfo *) plugin_info_list->data;
+    loaded_one |= grl_plugin_registry_load (registry, pinfo->filename, NULL);
+    plugin_info_list = g_list_next (plugin_info_list);
+  }
+
+  return loaded_one;
+}
+
+/* ================ PRIVATE API ================ */
+
+/**
+ * grl_plugin_registry_restrict_plugins:
+ * @registry: the registry instance
+ * @plugins: a @NULL-terminated array of plugins identifiers
+ *
+ * Restrict the plugins that application sees to this list.
+ *
+ * Other plugins will not be available for the application, unless it uses
+ * directly #grl_plugin_registry_load() function.
+ **/
+void
+grl_plugin_registry_restrict_plugins (GrlPluginRegistry *registry,
+                                      gchar **plugins)
+{
+  g_return_if_fail (GRL_IS_PLUGIN_REGISTRY (registry));
+  g_return_if_fail (plugins);
+
+  /* Free previous list */
+  if (registry->priv->allowed_plugins) {
+    g_slist_foreach (registry->priv->allowed_plugins, (GFunc) g_free, NULL);
+    g_slist_free (registry->priv->allowed_plugins);
+    registry->priv->allowed_plugins = NULL;
+  }
+
+  while (*plugins) {
+    registry->priv->allowed_plugins = g_slist_prepend (registry->priv->allowed_plugins,
+                                                       g_strdup (*plugins));
+    plugins++;
+  }
+}
+
+/* ================ PUBLIC API ================ */
 
 /**
  * grl_plugin_registry_get_default:
@@ -374,6 +565,7 @@ grl_plugin_registry_unregister_source (GrlPluginRegistry *registry,
     ret = FALSE;
   }
 
+  g_free (id);
   return ret;
 }
 
@@ -402,7 +594,7 @@ grl_plugin_registry_add_directory (GrlPluginRegistry *registry,
 /**
  * grl_plugin_registry_load:
  * @registry: the registry instance
- * @path: the path to the so file
+ * @library_filename: the path to the so file
  * @error: error return location or @NULL to ignore
  *
  * Loads a module from shared object file stored in @path
@@ -413,79 +605,124 @@ grl_plugin_registry_add_directory (GrlPluginRegistry *registry,
  */
 gboolean
 grl_plugin_registry_load (GrlPluginRegistry *registry,
-                          const gchar *path,
+                          const gchar *library_filename,
                           GError **error)
 {
   GModule *module;
   GrlPluginDescriptor *plugin;
+  GrlPluginInfo *plugin_info;
   GList *plugin_configs;
-  gchar *xml_path;
+  gchar *dirname;
+  gchar *plugin_info_filename;
+  gchar *plugin_info_fullpathname;
 
   g_return_val_if_fail (GRL_IS_PLUGIN_REGISTRY (registry), FALSE);
 
-  module = g_module_open (path, G_MODULE_BIND_LAZY);
+  module = g_module_open (library_filename, G_MODULE_BIND_LAZY);
   if (!module) {
-    GRL_WARNING ("Failed to open module: '%s'", path);
+    GRL_WARNING ("Failed to open module: '%s'", library_filename);
     g_set_error (error,
                  GRL_CORE_ERROR,
                  GRL_CORE_ERROR_LOAD_PLUGIN_FAILED,
-                 "Failed to load plugin at '%s'", path);
+                 "Failed to load plugin at '%s'", library_filename);
     return FALSE;
   }
 
   if (!g_module_symbol (module, "GRL_PLUGIN_DESCRIPTOR", (gpointer) &plugin)) {
-    GRL_WARNING ("Did not find plugin descriptor: '%s'", path);
+    GRL_WARNING ("Did not find plugin descriptor: '%s'", library_filename);
     g_set_error (error,
                  GRL_CORE_ERROR,
                  GRL_CORE_ERROR_LOAD_PLUGIN_FAILED,
-                 "'%s' is not a valid plugin file", path);
+                 "'%s' is not a valid plugin file", library_filename);
     g_module_close (module);
     return FALSE;
   }
 
   if (!plugin->plugin_init ||
-      !plugin->info.id) {
-    GRL_WARNING ("Plugin descriptor is not valid: '%s'", path);
+      !plugin->plugin_id) {
+    GRL_WARNING ("Plugin descriptor is not valid: '%s'", library_filename);
     g_set_error (error,
                  GRL_CORE_ERROR,
                  GRL_CORE_ERROR_LOAD_PLUGIN_FAILED,
-                 "'%s' is not a valid plugin file", path);
+                 "'%s' is not a valid plugin file", library_filename);
     g_module_close (module);
     return FALSE;
   }
 
-  plugin->info.filename = g_strdup (path);
+  /* Check if plugin is already loaded */
+  if (g_hash_table_lookup (registry->priv->plugins, plugin->plugin_id)) {
+    GRL_WARNING ("Plugin is already loaded: '%s'", library_filename);
+    g_set_error (error,
+                 GRL_CORE_ERROR,
+                 GRL_CORE_ERROR_LOAD_PLUGIN_FAILED,
+                 "'%s' is already loaded", library_filename);
+    g_module_close (module);
+    return FALSE;
+  }
 
-  xml_path = g_strconcat (GRL_PLUGINS_CONF_DIR,
-			  G_DIR_SEPARATOR_S,
-			  plugin->info.id,
-			  ".xml",
-			  NULL);
-  plugin->info.optional_info =  get_info_from_plugin_xml (xml_path);
-  g_free (xml_path);
+  plugin_info = g_hash_table_lookup (registry->priv->plugin_infos,
+                                     plugin->plugin_id);
 
-  set_plugin_rank (registry, plugin);
+  /* This happens when the user invokes directly this function: plugin
+     information has not been loaded yet */
+  if (!plugin_info) {
+    plugin_info_filename = g_strconcat (plugin->plugin_id,
+                                        "." GRL_PLUGIN_INFO_SUFFIX,
+                                        NULL);
+    dirname = g_path_get_dirname (library_filename);
+    plugin_info_fullpathname = g_build_filename (dirname,
+                                                 plugin_info_filename,
+                                                 NULL);
+    plugin_info = grl_plugin_registry_load_plugin_info (registry,
+                                                        plugin->plugin_id,
+                                                        plugin_info_filename);
+    if (!plugin_info) {
+      GRL_WARNING ("Plugin '%s' does not have XML information file '%s'",
+                   plugin->plugin_id,
+                   plugin_info_fullpathname);
+      /* Create a default one */
+      plugin_info = g_new0 (GrlPluginInfo, 1);
+      plugin_info->id = g_strdup (plugin->plugin_id);
+      plugin_info->filename = g_strdup (library_filename);
+      g_hash_table_insert (registry->priv->plugin_infos,
+                           plugin_info->id,
+                           plugin_info);
+      plugin_info->optional_info =
+        g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
+      g_hash_table_insert (plugin_info->optional_info,
+                           g_strdup (GRL_PLUGIN_INFO_MODULE),
+                           g_path_get_basename (plugin_info->filename));
+
+      set_plugin_rank (registry, plugin_info);
+    }
+    g_free (plugin_info_filename);
+    g_free (dirname);
+    g_free (plugin_info_fullpathname);
+  }
 
   g_hash_table_insert (registry->priv->plugins,
-		       (gpointer) plugin->info.id, plugin);
+                       (gpointer) plugin->plugin_id, plugin);
 
   plugin_configs = g_hash_table_lookup (registry->priv->configs,
-					plugin->info.id);
+                                        plugin->plugin_id);
 
-  if (!plugin->plugin_init (registry, &plugin->info, plugin_configs)) {
-    g_hash_table_remove (registry->priv->plugins, plugin->info.id);
-    GRL_WARNING ("Failed to initialize plugin: '%s'", path);
+  if (!plugin->plugin_init (registry, plugin_info, plugin_configs)) {
+    g_hash_table_remove (registry->priv->plugins, plugin->plugin_id);
+    GRL_INFO ("Failed to initialize plugin: '%s'", library_filename);
     g_set_error (error,
                  GRL_CORE_ERROR,
                  GRL_CORE_ERROR_LOAD_PLUGIN_FAILED,
-                 "Failed to initialize plugin at '%s'", path);
+                 "Failed to initialize plugin at '%s'", library_filename);
     g_module_close (module);
     return FALSE;
   }
 
+  /* Make plugin resident */
+  g_module_make_resident (module);
+
   plugin->module = module;
 
-  GRL_DEBUG ("Loaded plugin '%s' from '%s'", plugin->info.id, path);
+  GRL_DEBUG ("Loaded plugin '%s' from '%s'", plugin->plugin_id, library_filename);
 
   return TRUE;
 }
@@ -508,39 +745,21 @@ grl_plugin_registry_load_directory (GrlPluginRegistry *registry,
                                     const gchar *path,
                                     GError **error)
 {
-  GDir *dir;
-  gchar *file;
-  const gchar *entry;
-  gboolean loaded_one = FALSE;
+  GList *plugin_infos;
 
   g_return_val_if_fail (GRL_IS_PLUGIN_REGISTRY (registry), FALSE);
 
-  dir = g_dir_open (path, 0, NULL);
+  /* Load plugins information */
+  plugin_infos = grl_plugin_registry_load_plugin_info_directory (registry,
+                                                                 path,
+                                                                 error);
 
-  if (!dir) {
-    GRL_WARNING ("Could not open plugin directory: '%s'", path);
-    g_set_error (error,
-                 GRL_CORE_ERROR,
-                 GRL_CORE_ERROR_LOAD_PLUGIN_FAILED,
-                 "Failed to open plugin directory '%s'", path);
-    return FALSE;
-  }
-
-  while ((entry = g_dir_read_name (dir)) != NULL) {
-    if (g_str_has_suffix (entry, "." G_MODULE_SUFFIX)) {
-      file = g_build_filename (path, entry, NULL);
-      if (grl_plugin_registry_load (registry, file, NULL)) {
-        loaded_one = TRUE;
-      }
-      g_free (file);
-    }
-  }
-
-  if (!loaded_one) {
+  /* Load the plugins */
+  if (!grl_plugin_registry_load_list (registry, plugin_infos)) {
     GRL_WARNING ("No plugins loaded from directory '%s'", path);
   }
+  g_list_free (plugin_infos);
 
-  g_dir_close (dir);
   return TRUE;
 }
 
@@ -563,30 +782,82 @@ grl_plugin_registry_load_directory (GrlPluginRegistry *registry,
 gboolean
 grl_plugin_registry_load_all (GrlPluginRegistry *registry, GError **error)
 {
-  GSList *plugin_dir;
-  gboolean loaded_one = FALSE;
+  GList *all_plugin_infos;
+  gboolean loaded_one;
 
   g_return_val_if_fail (GRL_IS_PLUGIN_REGISTRY (registry), TRUE);
 
-  for (plugin_dir = registry->priv->plugins_dir;
-       plugin_dir;
-       plugin_dir = g_slist_next (plugin_dir)) {
-    if (grl_plugin_registry_load_directory (registry, plugin_dir->data, NULL)) {
-      loaded_one = TRUE;
-    }
+  /* Preload all plugin infos */
+  if (!registry->priv->all_plugin_info_loaded) {
+    grl_plugin_registry_load_plugin_info_all (registry);
+    registry->priv->all_plugin_info_loaded = TRUE;
   }
 
+  /* Now load all plugins */
+  all_plugin_infos = g_hash_table_get_values (registry->priv->plugin_infos);
+  loaded_one = grl_plugin_registry_load_list (registry, all_plugin_infos);
+
+  g_list_free (all_plugin_infos);
+
   if (!loaded_one) {
     g_set_error (error,
                  GRL_CORE_ERROR,
                  GRL_CORE_ERROR_LOAD_PLUGIN_FAILED,
-                 "All configured plugin paths are invalid. "    \
+                 "All configured plugin paths are invalid. "   \
                  "Failed to load plugins.");
   }
 
   return loaded_one;
 }
 
+/**
+ * grl_plugin_registry_load_by_id:
+ * @registry: the registry instance
+ * @plugin_id: plugin identifier
+ * @error: error return location or @NULL to ignore
+ *
+ * Loads plugin identified by @plugin_id.
+ *
+ * This requires the XML plugin information file to define a "module" key with
+ * the name of the module that provides the plugin or the absolute path of the
+ * actual module file.
+ *
+ * Returns: %TRUE if the plugin is loaded correctly
+ *
+ * Since: 0.1.14
+ **/
+gboolean
+grl_plugin_registry_load_by_id (GrlPluginRegistry *registry,
+                                const gchar *plugin_id,
+                                GError **error)
+{
+  GrlPluginInfo *plugin_info;
+
+  g_return_val_if_fail (GRL_IS_PLUGIN_REGISTRY (registry), FALSE);
+
+  /* Check if there is information loaded */
+  plugin_info = g_hash_table_lookup (registry->priv->plugin_infos, plugin_id);
+    /* Maybe we need to load all plugins before */
+  if (!plugin_info &&
+      !registry->priv->all_plugin_info_loaded) {
+    grl_plugin_registry_load_plugin_info_all (registry);
+    registry->priv->all_plugin_info_loaded = TRUE;
+    /* Search again */
+    plugin_info = g_hash_table_lookup (registry->priv->plugin_infos, plugin_id);
+  }
+
+  if (!plugin_info) {
+    GRL_WARNING ("There is no information about a plugin with id '%s'", plugin_id);
+    g_set_error (error,
+                 GRL_CORE_ERROR,
+                 GRL_CORE_ERROR_LOAD_PLUGIN_FAILED,
+                 "There is no information about a plugin with id '%s'", plugin_id);
+    return FALSE;
+  }
+
+  return grl_plugin_registry_load (registry, plugin_info->filename, error);
+}
+
 /**
  * grl_plugin_registry_lookup_source:
  * @registry: the registry instance
@@ -699,7 +970,7 @@ grl_plugin_registry_get_sources_by_operations (GrlPluginRegistry *registry,
  * Unload from memory a module identified by @plugin_id. This means call the
  * module's deinit function.
  *
- * Returns %TRUE% on success.
+ * Returns: %TRUE% on success.
  *
  * Since: 0.1.7
  */
@@ -751,10 +1022,7 @@ grl_plugin_registry_unload (GrlPluginRegistry *registry,
     plugin->plugin_deinit ();
   }
 
-  g_free (plugin->info.filename);
-  if (plugin->info.optional_info) {
-    g_hash_table_destroy (plugin->info.optional_info);
-  }
+  g_hash_table_remove (registry->priv->plugins, plugin_id);
 
   if (plugin->module) {
     g_module_close (plugin->module);
@@ -895,8 +1163,8 @@ grl_plugin_registry_lookup_metadata_key (GrlPluginRegistry *registry,
  *
  * @key is included in that list.
  *
- * Returns: (element-type Grl.KeyID) (transfer-none): a #GList of related
- * keys, or @NULL if key is invalid.
+ * Returns: (element-type GObject.ParamSpec) (transfer none): a #GList of
+ * related keys, or @NULL if key is invalid.
  *
  * Since: 0.1.10
  **/
@@ -981,6 +1249,7 @@ grl_plugin_registry_add_config (GrlPluginRegistry *registry,
     /* Notice that we are using g_list_append on purpose to avoid
        having to insert again in the hash table */
     configs = g_list_append (configs, config);
+    g_free (plugin_id);
   } else {
     configs = g_list_prepend (configs, config);
     g_hash_table_insert (registry->priv->configs,
